import QUB
from Buttons import *
from qubtree import Node
from math import *

# class ModelTree: manage a qubtree.Node model representation
# this is _so_ not valid unless it's the sole manipulator of the Node in question

def RateConnects(rate, s1, s2):
    states = rate['States'].data
    return ( (states[0] == s1 and states[1] == s2)
          or (states[0] == s2 and states[1] == s1) )

class ModelTree:
    def __init__(self, node=None):
        if not node:
            node = Node('ModelFile')
            node['ChannelCount'].data = 1
            node['Amps'].data = range(0,10)
            node['Stds'].data = [0.1] * 10
            node['NAr'].data = [0] * 10
            ars = node['Ars']
            for i in range(10):
                ars.insert('Ar').data = [0.0] * 10
                
        self.node = node
        self.buildStates()
        self.buildRates()

    def buildStates(self):
        states = []
        classesPresent = []
        state = self.node['States'].find('State')
        while not state.isNull:
            states.append(state)

            cls = state['Class'].data[0]
            if not cls in classesPresent:
                classesPresent.append(cls)
            
            state = state.nextSameName()
            
        self.states = states
        self.nstate = len(states)
        self.classesPresent = classesPresent

    def buildRates(self):
        rates = []
        rate = self.node['Rates'].find('Rate')
        while not rate.isNull:
            rates.append(rate)
            rate = rate.nextSameName()
        self.rates = rates

    def addState(self, cls):
        state = Node('State')
        state['Class'].data = cls
        state['Pr'].data = 0.0
        state['x'].data = 0
        state['y'].data = 0
        
        self.node['States'].append(state)
        self.states.append(state)
        self.nstate = self.nstate + 1

        return state

    def addRate(self, s1, s2):
        rate = QUB.ModelView.makeNewRate()
        # rate = self.node['Rates'][0].clone() # must change since there may be no rates avail. to clone
        rate['States'].data = [s1, s2]
        self.node['Rates'].append(rate)
        self.rates.append(rate)

        return len(self.rates) - 1        

    def addStateFrom(self, s1, cls):
        state = self.addState(cls)
        self.addRate(s1, self.nstate-1)
        return state

    def addStateBetween(self, s1, s2, cls):
        try:
            rate_ix = [i for i in range(len(self.rates)) if RateConnects(self.rates[i], s1, s2)][0]
            self.removeRate(rate_ix)
        except:
            pass # no existing rate to remove; no matter
        state = self.addState(cls)
        state_ix = self.nstate - 1
        state['x'].data = (self.states[s1]['x'].data[0] + self.states[s2]['x'].data[0]) / 2
        state['y'].data = (self.states[s1]['y'].data[0] + self.states[s2]['y'].data[0]) / 2
        self.addRate(s1, state_ix)
        self.addRate(state_ix, s2)
        return state

    def removeRate(self, i):
        self.node['Rates'].remove( self.rates[i] )
        del self.rates[i]

    def removeState(self, i):
        self.node['States'].remove( self.states[i] )
        del self.states[i]
        self.nstate = self.nstate - 1

# StyledModelBuilder:
#   keep on adding states until the MIL likelihood levels out
#   do this for each class present in the initial model
#   AddState is a function(ModelTree, clsOfStateToAdd, props)
#   -- provide a different version to choose e.g. star or chain topology
        
def StyledModelBuilder(props, AddState):
    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return
    model = ModelTree( modelObj.getAsNode() )
    if model.nstate == 0:
        QUB.Report('no states in the model')
        return

    optimizer = props['optimizer']    
    deltaLL = props['deltaLL']
    newSettings = Node()
    newSettings['Mode'].data = 'optimize'
    newSettings['use segments'].data = 'together'
    newSettings['ShowIterations'].data = 0

    Run( 'Modeling', optimizer, newSettings )

    postLL = QUB.ResultsView.currResults.node['LL'].data[0]
    preLL = postLL - 2 * deltaLL # damn you lack of do...while
    
    for cls in model.classesPresent:
        newrate_ix = -1
        newstate_ix = -1
        while (postLL - preLL) > deltaLL:
            lastGoodModel = ModelTree( modelObj.getAsNode() )
            model = ModelTree( modelObj.getAsNode() )
            AddState(model, cls, props)
            modelObj.setAsNode( model.node )
            
            preLL = postLL
            Run( 'Modeling', optimizer, newSettings )
            postLL = QUB.ResultsView.currResults.node['LL'].data[0]

        modelObj.setAsNode( lastGoodModel.node )
        preLL = postLL - 2 * deltaLL            

    Run( 'Modeling', optimizer, newSettings )


# instantiate a StyledModelBuilder function with the chosen topology, suitable for AddTaskButton
def ModelBuilder(AddState):
    return lambda props: StyledModelBuilder(props, AddState)
        
# star topology

def AngleOfStates(s1, s2):
    dx = s2['x'].data[0] - s1['x'].data[0]
    dy = s2['y'].data[0] - s1['y'].data[0]
    mag = sqrt(dx*dx + dy*dy)
    angle = asin(dy/mag)
    if dx < 0:
        if dy >= 0:
            angle = pi - angle
        else:
            angle = -pi - angle
    return angle

def AddStarState(model, cls, props):
    hub_ix = props['hubState'] - 1
    if hub_ix >= len(model.states):
        hub_ix = len(model.states) - 1
    hubstate = model.states[ hub_ix ]
    
    angles = [ (AngleOfStates(hubstate, model.states[i]), i) for i in range(model.nstate) if i != hub_ix ]
    angles.sort()
    angles.append( (angles[0][0] + 2*pi, angles[0][1]) )
    #              (ang. dist. from sorted_i to +1, state_ix, sorted state_ix)
    anglediffs = [ (angles[i+1][0] - angles[i][0], angles[i][1], i) for i in range(len(angles) - 1) ]
    anglediffs.sort()
    insertafter = max(anglediffs)
    insertaftersorted = insertafter[2]
    abef, aaft = angles[ insertaftersorted ][0], angles[ insertaftersorted + 1 ][0]
    insertangle =  abef + (aaft - abef) / 2
    if insertangle >= 2*pi:
        insertangle = insertangle - 2*pi
    x0, y0 = hubstate['x'].data[0], hubstate['y'].data[0]
    dx, dy = cos(insertangle), sin(insertangle)
    if dx != 0:
        xmag = max( (100 - x0) / dx, -x0 / dx )
    else:
        xmag = 1.0e10
    if dy != 0:
        ymag = max( (100 - y0) / dy, -y0 / dy )
    else:
        ymag = 1.0e10
    mag = 0.8 * min( xmag, ymag )

    state = model.addStateFrom(hub_ix, cls)    
    state['x'].data = x0 + mag * dx
    state['y'].data = y0 + mag * dy

AddTaskButton( 'MakeStar', # button text
               ModelBuilder(AddStarState),   # function
               'Star',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Build up a star model until LL stops improving",   # tool-tip
               "Keep on adding states in a star configuration until the MIL likelihood levels out.", # description for the "Edit Buttons" dialog
               "MakeStar properties:", # caption for the settings dialog
               [TextField("deltaLL", "Delta LL:", "0.5", acceptFloatGreaterThan(0)),
                DialogRadioGroup("optimizer", "Optimizer:", ['MIL', 'MPL', 'MSL', 'Mac'], 'MIL'),
                TextField("hubState", "Starting state:", "1", acceptIntGreaterThan(0))],
               'Scripts',
               DataChannel=True, DataList=True, FileList=True)

# chain topology

def Connections(model):
    connections = {}
    for rate in model.rates:
        states = rate['States'].data
        if connections.has_key(states[0]):
            connections[ states[0] ].append( states[1] )
        else:
            connections[ states[0] ] = [ states[1] ]
        if connections.has_key(states[1]):
            connections[ states[1] ].append( states[0] )
        else:
            connections[ states[1] ] = [ states[0] ]
    return connections

def LeftmostState(model):
    return min( [(model.states[i]['x'].data[0], i) for i in range(model.nstate)] )[1]

def ChainStateOrder(model):
    conn = Connections(model)
    at = LeftmostState(model)
    order = [at]
    while len(conn[at]):
        next = conn[at][0]
        del conn[at][0]
        conn[next].remove(at)
        at = next
        order.append(at)
    return order

def SubChain(model, cls):
    order = ChainStateOrder(model)
    subchain = [(order[i], i) for i in range(len(order)) if model.states[order[i]]['Class'].data[0] == cls] # assumes all states of cls are contiguous in model
    prev = None
    post = None

    ixInOrder = subchain[0][1] # index of first subchain item in order
    if ixInOrder > 0:
        prev = order[ ixInOrder - 1 ]
    ixInOrder = subchain[ len(subchain) - 1 ][1]
    if ixInOrder < (len(order) - 1):
        post = order[ ixInOrder + 1 ]

    subchain = [item[0] for item in subchain]
    return prev, subchain, post

def AddChainState(model, cls, props):
    prev, statesOfCls, post = SubChain(model, cls)

    if len(statesOfCls) == 1 and prev == None:
        state = model.addStateFrom( statesOfCls[0], cls )
        state['x'].data = 5
        state['y'].data = model.states[ statesOfCls[0] ]['y'].data
    elif len(statesOfCls) == 1 and post == None:
        state = model.addStateFrom( statesOfCls[0], cls )
        state['x'].data = 95
        state['y'].data = model.states[ statesOfCls[0] ]['y'].data
    else:
        if post: statesOfCls.append(post)
        gaps = [ (model.states[ statesOfCls[i+1] ]['x'].data[0] - model.states[ statesOfCls[i] ]['x'].data[0], i) for i in range(len(statesOfCls)-1) ]
        gaps.sort()
        after = gaps[ len(gaps) - 1 ][1]
        model.addStateBetween(statesOfCls[after], statesOfCls[after+1], cls)

test = AddTaskButton( 'MakeChain', # button text
               ModelBuilder(AddChainState),   # function
               'Chain',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "2 Build a chain model as long as LL improves",   # tool-tip
               "Keep on adding states in a chain configuration until the MIL likelihood levels out.", # description for the "Edit Buttons" dialog
               "MakeChain properties:", # caption for the settings dialog
               [TextField("deltaLL", "Delta LL:", "0.5", acceptFloatGreaterThan(0)),
                DialogRadioGroup("optimizer", "Optimizer:", ['MIL', 'MPL', 'MSL', 'Mac'], 'MIL'),
],
               'Scripts',
               DataChannel=True, DataList=True, FileList=True)
           
